# AUTOGENERATED! DO NOT EDIT! File to edit: 01_foregrounds.ipynb (unless otherwise specified).

__all__ = ['emission_naming', 'load_sky_emission']

# Cell

import healpy as hp
import numpy as np

from pathlib import Path
import logging as log

from .core import get_telescope, base_folder

# Cell

emission_naming = dict(
    foreground_emission="foreground",
    CMB_unlensed="cmb-unlensed",
    CMB_lensing_signal="cmb-lensing",
    CMB_tensor_to_scalar_ratio="cmb-tensors",
)

# Cell


def load_sky_emission(config_sky_emission, site, channel):
    """Load foreground maps for a channel

    Parameters
    ----------
    config_sky_emission : dict
        CMB-S4 sky emission configuration,
        generally config['sky_emission'] for a configuration
        loaded from a TOML file
        dictionary with standard emission names and their weights
    site : str
        'Pole' or 'Chile', case doesn't matter
    channel : str
        Channel tag, e.g. 'MFHS1'

    Returns
    -------
    output_map : numpy array
        Output map with all emissions combined, uses nan for missing pixels
    """
    telescope = get_telescope(channel)
    nside = 512 if telescope == "SAT" else 4096
    npix = hp.nside2npix(nside)
    output_map = np.zeros((3, npix), dtype=np.float32)
    realization = 0  # foregrounds are deterministic
    for emission, weight in config_sky_emission.items():
        if weight == 0:
            log.info("Skip %s", emission)
            continue
        log.info("Processing %s", emission)

        emission_map = hp.read_map(
            Path(base_folder)
            / f"{realization:08d}"
            / f"{site.lower()}_{emission_naming[emission]}_{telescope}_{channel}_filtered_telescope_all_time_all_bmap.fits",
            (0, 1, 2),
            verbose=False,
        )
        emission_map[emission_map == hp.UNSEEN] = np.nan
        if emission == "CMB_tensor_to_scalar_ratio":
            weight /= 0.01  # r of the input maps
        output_map += emission_map * weight

    output_map /= 1e6  # uK -> K
    return output_map


# AUTOGENERATED! DO NOT EDIT! File to edit: 02_atmosphere.ipynb (unless otherwise specified).

__all__ = ['get_telecope_years', 'load_atmosphere']

# Cell

import healpy as hp
import numpy as np
from pathlib import Path
import logging as log

from .core import get_telescope, base_folder, simulations_observing_efficiency

# Cell


def get_telecope_years(config, site, channel):
    """Compute the number of telescope/years in the CMB-S4 configuration

    config_telescopes : dict
        CMB-S4 telescopes configuration,
        generally loaded from a TOML file
    site : str
        'Pole' or 'Chile', case doesn't matter
    channel : str
        Channel tag, e.g. 'MFHS1'
    """
    telescope_years = 0
    for tube, bands in config["telescopes"][get_telescope(channel)].items():
        if bands["site"].lower() == site.lower():
            has_band = bands.get(channel[:-1], 0) > 0
            telescope_years += has_band * bands.get(
                "years", config["experiment"]["total_experiment_length_years"]
            )
    return telescope_years

# Cell


def load_atmosphere(config, site, channel, realization=0):
    """Load foreground maps for a channel

    Parameters
    ----------
    config : dict
        CMB-S4 configuration,
        generally loaded from a TOML file
    site : str
        'Pole' or 'Chile', case doesn't matter
    channel : str
        Channel tag, e.g. 'MFHS1'
    realization : int
        Choose one of the available 8 realizations

    Returns
    -------
    output_map : numpy array
        Output map with all emissions combined, uses nan for missing pixels
    """

    telescope = get_telescope(channel)
    map_filename = (
        Path(f"{realization:08d}")
        / f"{site.lower()}_atmosphere_{telescope}_{channel}_filtered_telescope_all_time_all_bmap.fits"
    )
    log.info(f"Reading {map_filename}")
    atmosphere_map = hp.read_map(
        Path(base_folder) / map_filename, (0, 1, 2), dtype=np.float32, verbose=False
    )
    atmosphere_map[atmosphere_map == hp.UNSEEN] = np.nan
    # input map is 10 days at 100% efficiency
    atmosphere_map *= np.sqrt(10) / np.sqrt(
        365.25 * config["experiment"]["observing_efficiency"]
    )
    # pole sims have lower efficiency
    atmosphere_map *= np.sqrt(
        simulations_observing_efficiency[site.lower()].get(telescope, 1)
    )
    atmosphere_map /= np.sqrt(get_telecope_years(config, site, channel))
    return atmosphere_map


# AUTOGENERATED! DO NOT EDIT! File to edit: 04_cli.ipynb (unless otherwise specified).

__all__ = ['md5sum_string', 'md5sum_file', 'S4RefSimTool', 'command_line_script']

# Cell

import toml
import healpy as hp
import numpy as np
import h5py
from pathlib import Path
import logging as log
from datetime import date

from .core import get_telescope, parse_channels

from s4_design_sim_tool import __version__

from .foregrounds import load_sky_emission
from .atmosphere import load_atmosphere, get_telecope_years
from .noise import load_noise

# Cell

import hashlib

def md5sum_string(string):
    return hashlib.md5(string.encode("utf-8")).hexdigest()

def md5sum_file(filename):
    """Compute md5 checksum of the contents of a file"""
    return md5sum_string(open(filename,'r').read())

# Cell


class S4RefSimTool:
    def __init__(self, config_filename, output_folder="output"):
        """Simulate CMB-S4 maps based on the experiment configuration

        Parameters
        ----------
        config : filename
            CMB-S4 configuration stored in a TOML file
            see for example s4_reference_design.toml in the repository
        output_folder : str or Path
            Output path
        """
        self.config_filename = config_filename
        self.config = toml.load(self.config_filename)
        self.output_filename_template = (
            "cmbs4_{tag}_KCMB_{telescope}-{band}_nside{nside}_{split}_of_{nsplits}.fits"
        )
        self.output_folder = Path(output_folder)
        self.output_folder.mkdir(parents=True, exist_ok=True)

    def run(self, channels="all", sites=["Pole", "Chile"], write_outputs=True):
        """Run the simulation

        Parameters
        ----------
        channels : str or list[str]
            list of channel tags, e.g.
            * ["LFS1", "LFS2"] or
            * "SAT" or "LAT"
            * "all" (default)
        site : list[str]
            ['Pole'] or ['Chile'], by default ["Pole", "Chile"]
        write_outputs : bool
            if True write the outputs to disk, if False return them (not implemented yet)
        """
        nsplits = self.config["experiment"].get("number_of_splits", 1)
        assert (
            nsplits < 8
        ), "We currently only have 7 independent realizations of atmosphere and noise"
        for site in sites:
            for channel in parse_channels(channels):
                if get_telecope_years(self.config, site, channel) == 0:
                    continue
                log.info(f"Simulate channel {channel} at {site}")
                sky_emission = load_sky_emission(
                    self.config["sky_emission"], site, channel
                )
                for split in range(nsplits):
                    output_map = load_atmosphere(
                        self.config, site, channel, realization=split
                    )
                    output_map += load_noise(
                        self.config, site, channel, realization=split
                    )
                    if split > 0:
                        output_map *= np.sqrt(nsplits)
                    output_map += sky_emission
                    if write_outputs:
                        output_filename = self.output_filename_template.format(
                            nside=hp.npix2nside(len(output_map[0])),
                            telescope=get_telescope(channel),
                            band=channel,
                            tag="sky_atmosphere_noise",
                            split=max(
                                1, split
                            ),  # split=0 is full mission and we want 1
                            nsplits=nsplits,
                        )
                        log.info(f"Writing {output_filename}")
                        hp.write_map(
                            self.output_folder / output_filename,
                            output_map,
                            column_units="K_CMB",
                            extra_header=[
                                ("SOFTWARE", "s4_design_sim_tool"),
                                ("SW_VERS", __version__),
                                ("SKY_VERS", "1.0"),
                                ("ATM_VERS", "1.0"),
                                ("NOI_VERS", "1.0"),
                                ("DATE", str(date.today())),
                                ("CONFMD5", md5sum_file(self.config_filename)),
                            ],
                            coord="Q",
                            overwrite=True,
                        )
                    else:
                        raise NotImplementedError("Only writing FITS output for now")

# Cell


def command_line_script(args=None):

    import logging as log

    log.basicConfig(level=log.INFO)

    import argparse

    parser = argparse.ArgumentParser(description="Run s4_design_sim_tool")
    parser.add_argument("config", type=str, help="Configuration file")
    parser.add_argument(
        "--channels",
        type=str,
        help="Channels e.g. all, SAT, LAT, LFL1 or comma separated list of channels",
        required=False,
        default="all",
    )
    parser.add_argument(
        "--site",
        type=str,
        help="Pole, Chile or all, default all",
        required=False,
        default="all",
    )
    parser.add_argument(
        "--output_folder",
        type=str,
        help="Output folder, optional",
        required=False,
        default="output",
    )
    res = parser.parse_args(args)
    if res.site == "all":
        sites = ["Chile", "Pole"]
    else:
        sites = [res.site]
    sim = S4RefSimTool("s4_reference_design.toml", output_folder=res.output_folder)
    sim.run(channels=res.channels, sites=sites)


# AUTOGENERATED! DO NOT EDIT! File to edit: 00_core.ipynb (unless otherwise specified).

__all__ = ['base_folder', 'simulations_observing_efficiency', 'get_telescope', 'parse_channels', 's4']

# Cell

import os
base_folder = os.environ.get("S4REFSIMTOOL_INPUT",
                             "/global/cscratch1/sd/keskital/s4sim/reference_tool/out/")

# Cell

simulations_observing_efficiency = dict(
    pole=dict(SAT=46.29 / 100, LAT=37.23 / 100), chile=dict()
)

# Cell


def get_telescope(channel):
    """Identify if a channel is in SAT or LAT

    Parameters
    ----------
    channel : str
        channel tag, e.g. MFH1

    Returns
    -------
    telescope : str
        'SAT' or 'LAT'
    """
    return channel[-2] + "AT"

# Cell

import h5py

s4 = h5py.File("cmbs4_tophat.h5", "r")


def parse_channels(channels):
    if channels in ["SAT", "LAT"]:
        channels = [ch for ch in s4.keys() if s4[ch].attrs["telescope"] == channels]
    elif channels in ["all", None]:
        channels = list(s4.keys())
    elif isinstance(channels, str):
        channels = channels.split(",")
    return channels


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"base_folder": "00_core.ipynb",
         "simulations_observing_efficiency": "00_core.ipynb",
         "get_telescope": "00_core.ipynb",
         "parse_channels": "00_core.ipynb",
         "s4": "00_core.ipynb",
         "emission_naming": "01_foregrounds.ipynb",
         "load_sky_emission": "01_foregrounds.ipynb",
         "get_telecope_years": "02_atmosphere.ipynb",
         "load_atmosphere": "02_atmosphere.ipynb",
         "thinfp_string": "03_noise.ipynb",
         "thinfp_table": "03_noise.ipynb",
         "get_thinfp": "03_noise.ipynb",
         "get_tube_years": "03_noise.ipynb",
         "load_noise": "03_noise.ipynb",
         "md5sum_string": "04_cli.ipynb",
         "md5sum_file": "04_cli.ipynb",
         "S4RefSimTool": "04_cli.ipynb",
         "command_line_script": "04_cli.ipynb"}

modules = ["core.py",
           "foregrounds.py",
           "atmosphere.py",
           "noise.py",
           "cli.py"]

doc_url = "https://cmb-s4.github.io/s4_design_sim_tool/"

git_url = "https://github.com/cmb-s4/s4_design_sim_tool/tree/master/"

def custom_doc_links(name): return None
